
from enum import IntEnum

class MetadataType(IntEnum):
    Title = 0
    Artist = 1
    Creator = 2
    Version = 3
    Source = 4
    Tags = 5
    VideoDataOffset = 6
    VideoDataLength = 7
    VideoHash = 8
    BeatmapSetID = 9
    Genre = 10
    Language = 11
    TitleUnicode = 12
    ArtistUnicode = 13
    Difficulty = 14
    PreviewTime = 15
    ArtistFullName = 16
    ArtistTwitter = 17
    SourceUnicode = 18
    ArtistURL = 19
    Revision = 20
    PackID = 21
    Unknown = 9999




from .utils import uint32_slice_to_byte_slice
from typing import List
from numba import njit
import numpy as np

class SimpleCryptor:
    def __init__(self, key: List[int]) -> None:
        # Pre-compute byte key for better performance
        self.key = np.array(uint32_slice_to_byte_slice(key), dtype=np.uint8)

    def encrypt_bytes(self, buf: bytearray) -> None:
        buf_array = np.frombuffer(buf, dtype=np.uint8)
        _encrypt_bytes_jit(buf_array, self.key)
        buf[:] = buf_array.tobytes()

    def decrypt_bytes(self, buf: bytearray) -> None:
        buf_array = np.frombuffer(buf, dtype=np.uint8)
        _decrypt_bytes_jit(buf_array, self.key)
        buf[:] = buf_array.tobytes()

@njit(cache=True, inline='always')
def _rotate_left(val: int, n: int) -> int:
    val &= 0xFF
    n &= 0x07
    return ((val << n) | (val >> (8 - n))) & 0xFF

@njit(cache=True, inline='always')
def _rotate_right(val: int, n: int) -> int:
    val &= 0xFF
    n &= 0x07
    return ((val >> n) | (val << (8 - n))) & 0xFF

@njit(cache=True)
def _encrypt_bytes_jit(buf: np.ndarray, byte_key: np.ndarray) -> None:
    prev_encrypted = np.uint8(0)
    buf_len = len(buf)

    for i in range(buf_len):
        sum_val = buf[i] + (byte_key[i % 16] >> 2)
        buf[i] = ((sum_val % 256) + 256) % 256
        buf[i] ^= _rotate_left(byte_key[15 - i % 16], (prev_encrypted + buf_len - i) % 7)
        buf[i] = _rotate_right(buf[i], (~prev_encrypted & 0xFFFFFFFF) % 7)
        prev_encrypted = buf[i]

@njit(cache=True)
def _decrypt_bytes_jit(buf: np.ndarray, byte_key: np.ndarray) -> None:
    prev_encrypted = np.uint8(0)
    buf_len = len(buf)

    for i in range(buf_len):
        tmp_e = buf[i]
        buf[i] = _rotate_left(buf[i], (~prev_encrypted & 0xFFFFFFFF) % 7)
        buf[i] ^= _rotate_left(byte_key[15 - i % 16], (prev_encrypted + buf_len - i) % 7)
        diff = buf[i] - (byte_key[i % 16] >> 2)
        buf[i] = ((diff % 256) + 256) % 256
        prev_encrypted = tmp_e




from .simple_cryptor import SimpleCryptor
from typing import List
import struct

TEA_DELTA = 0x9E3779B9
TEA_ROUNDS = 32

class XTEA:
    """XTEA implements the Extended Tiny Encryption Algorithm"""

    def __init__(self, key: List[int]) -> None:
        self.key = key
        self.simple_cryptor = SimpleCryptor(key)

    def decrypt(self, buffer: bytearray, start: int, count: int) -> None:
        self._encrypt_decrypt(buffer, start, count, False)

    def encrypt(self, buffer: bytearray, start: int, count: int) -> None:
        self._encrypt_decrypt(buffer, start, count, True)

    def _encrypt_decrypt(self, buffer: bytearray, buf_start: int, count: int, encrypt: bool) -> None:
        full_word_count = count // 8
        left_over = count % 8

        # Process full 8-byte words
        for i in range(full_word_count):
            offset = buf_start + i * 8
            v0 = struct.unpack_from('<I', buffer, offset)[0]
            v1 = struct.unpack_from('<I', buffer, offset + 4)[0]

            if encrypt:
                v0, v1 = self._encrypt_word(v0, v1)
            else:
                v0, v1 = self._decrypt_word(v0, v1)

            struct.pack_into('<I', buffer, offset, v0)
            struct.pack_into('<I', buffer, offset + 4, v1)

        # Handle leftover bytes
        if left_over > 0:
            leftover_start = buf_start + full_word_count * 8
            leftover_buf = buffer[leftover_start:leftover_start + left_over]
            
            if encrypt:
                self.simple_cryptor.encrypt_bytes(leftover_buf)
            else:
                self.simple_cryptor.decrypt_bytes(leftover_buf)
            
            buffer[leftover_start:leftover_start + left_over] = leftover_buf

    def _encrypt_word(self, v0: int, v1: int) -> tuple:
        sum_val = 0
        for _ in range(TEA_ROUNDS):
            v0 = (v0 + ((((v1 << 4) ^ (v1 >> 5)) + v1) ^ (sum_val + self.key[sum_val & 3]))) & 0xFFFFFFFF
            sum_val = (sum_val + TEA_DELTA) & 0xFFFFFFFF
            v1 = (v1 + ((((v0 << 4) ^ (v0 >> 5)) + v0) ^ (sum_val + self.key[(sum_val >> 11) & 3]))) & 0xFFFFFFFF
        return v0, v1

    def _decrypt_word(self, v0: int, v1: int) -> tuple:
        # Calculate sum with proper overflow handling
        sum_val = 0
        for _ in range(TEA_ROUNDS):
            sum_val = (sum_val + TEA_DELTA) & 0xFFFFFFFF

        for _ in range(TEA_ROUNDS):
            v1 = (v1 - ((((v0 << 4) ^ (v0 >> 5)) + v0) ^ (sum_val + self.key[(sum_val >> 11) & 3]))) & 0xFFFFFFFF
            sum_val = (sum_val - TEA_DELTA) & 0xFFFFFFFF
            v0 = (v0 - ((((v1 << 4) ^ (v1 >> 5)) + v1) ^ (sum_val + self.key[sum_val & 3]))) & 0xFFFFFFFF
        return v0, v1




from .simple_cryptor import SimpleCryptor
from numba import njit, prange
from typing import List
import numpy as np

MAX_WORDS = 16
MAX_BYTES = MAX_WORDS * 4
TEA_DELTA = 0x9E3779B9

class XXTEA:
    """XXTEA implements the Corrected Block TEA algorithm"""

    def __init__(self, key: List[int]) -> None:
        self.cryptor = SimpleCryptor(key)
        self.key = np.array(key, dtype=np.uint32)

        # Pre-compute all possible key permutations for faster lookup
        self.key_table = np.array([[key[i ^ e] for i in range(4)] for e in range(4)], dtype=np.uint32)
        self.n = 0

    def decrypt(self, buffer: bytearray, start: int, count: int) -> None:
        self.encrypt_decrypt(buffer, start, count, False)

    def encrypt(self, buffer: bytearray, start: int, count: int) -> None:
        self.encrypt_decrypt(buffer, start, count, True)

    def encrypt_decrypt(self, buffer: bytearray, buf_start: int, count: int, encrypt: bool) -> None:
        full_word_count = count // MAX_BYTES
        left_over = count % MAX_BYTES

        # Process full blocks
        if full_word_count > 0:
            if encrypt:
                self.encrypt_full_blocks(buffer, buf_start, full_word_count)
            else:
                self.decrypt_full_blocks(buffer, buf_start, full_word_count)

        if left_over == 0:
            return

        leftover_start = buf_start + full_word_count * MAX_BYTES
        self.n = left_over // 4

        if self.n > 1:
            if encrypt:
                self.encrypt_words(self.n, self.key, buffer, leftover_start)
            else:
                self.decrypt_words(self.n, self.key, buffer, leftover_start)

            left_over -= self.n * 4
            if left_over == 0:
                return

            leftover_start += self.n * 4

        remaining = buffer[leftover_start:leftover_start + left_over]

        if encrypt:
            self.cryptor.encrypt_bytes(remaining)
        else:
            self.cryptor.decrypt_bytes(remaining)

        buffer[leftover_start:leftover_start + left_over] = remaining

    def encrypt_full_blocks(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Use parallel processing for multiple blocks
        if full_word_count >= 4:
            self.encrypt_full_blocks_parallel(buffer, buf_start, full_word_count)
            return

        # Sequential for small data
        for i in range(full_word_count):
            offset = buf_start + i * MAX_BYTES
            self.encrypt_fixed_word_array(self.key, buffer, offset)

    def decrypt_full_blocks(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Use parallel processing for multiple blocks
        if full_word_count >= 4:
            self.decrypt_full_blocks_parallel(buffer, buf_start, full_word_count)
            return

        # Sequential for small data
        for i in range(full_word_count):
            offset = buf_start + i * MAX_BYTES
            self.decrypt_fixed_word_array(self.key, buffer, offset)

    def encrypt_full_blocks_parallel(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Convert buffer slice to numpy array for parallel processing
        buffer_size = full_word_count * MAX_BYTES
        data = np.frombuffer(buffer[buf_start:buf_start + buffer_size], dtype=np.uint32).copy()
        data = _encrypt_blocks_parallel(data, self.key, full_word_count)
        buffer[buf_start:buf_start + buffer_size] = data.tobytes()

    def decrypt_full_blocks_parallel(self, buffer: bytearray, buf_start: int, full_word_count: int) -> None:
        # Convert buffer slice to numpy array for parallel processing
        buffer_size = full_word_count * MAX_BYTES
        data = np.frombuffer(buffer[buf_start:buf_start + buffer_size], dtype=np.uint32).copy()
        data = _decrypt_blocks_parallel(data, self.key, full_word_count)
        buffer[buf_start:buf_start + buffer_size] = data.tobytes()

    @staticmethod
    def encrypt_words(n: int, key: np.ndarray, data: bytearray, offset: int) -> None:
        v = np.frombuffer(data[offset:offset + n*4], dtype=np.uint32).copy()
        v = _encrypt_block(v, key, n)
        data[offset:offset + n*4] = v.tobytes()

    @staticmethod
    def decrypt_words(n: int, key: np.ndarray, data: bytearray, offset: int) -> None:
        v = np.frombuffer(data[offset:offset + n*4], dtype=np.uint32).copy()
        v = _decrypt_block(v, key, n)
        data[offset:offset + n*4] = v.tobytes()

    @staticmethod
    def encrypt_fixed_word_array(key: np.ndarray, data: bytearray, offset: int) -> None:
        if len(data) - offset < MAX_BYTES:
            return

        v = np.frombuffer(data[offset:offset + MAX_BYTES], dtype=np.uint32).copy()
        v = _encrypt_block_fixed(v, key)
        data[offset:offset + MAX_BYTES] = v.tobytes()

    @staticmethod
    def decrypt_fixed_word_array(key: np.ndarray, data: bytearray, offset: int) -> None:
        if len(data) - offset < MAX_BYTES:
            return

        v = np.frombuffer(data[offset:offset + MAX_BYTES], dtype=np.uint32).copy()
        v = _decrypt_block_fixed(v, key)
        data[offset:offset + MAX_BYTES] = v.tobytes()

@njit(cache=True, inline='always')
def _mx(y: int, z: int, sum_val: int, key_val: int) -> int:
    """The MX function used in XXTEA algorithm"""
    return ((((z >> 5) ^ (y << 2)) + ((y >> 3) ^ (z << 4))) ^ 
            ((sum_val ^ y) + (key_val ^ z))) & 0xFFFFFFFF

@njit(cache=True)
def _encrypt_block(v: np.ndarray, key: np.ndarray, n: int) -> np.ndarray:
    """JIT-compiled encryption logic for variable-size blocks"""
    rounds = 6 + 52 // n
    sum_val = np.uint32(0)
    z = v[n - 1]

    for _ in range(rounds):
        sum_val = (sum_val + TEA_DELTA) & 0xFFFFFFFF
        e = (sum_val >> 2) & 3

        for p in range(n - 1):
            y = v[p + 1]
            mx_val = _mx(y, z, sum_val, key[(p & 3) ^ e])
            v[p] = (v[p] + mx_val) & 0xFFFFFFFF
            z = v[p]

        y = v[0]
        mx_val = _mx(y, z, sum_val, key[((n - 1) & 3) ^ e])
        v[n - 1] = (v[n - 1] + mx_val) & 0xFFFFFFFF
        z = v[n - 1]

    return v

@njit(cache=True)
def _decrypt_block(v: np.ndarray, key: np.ndarray, n: int) -> np.ndarray:
    """JIT-compiled decryption logic for variable-size blocks"""
    rounds = 6 + 52 // n
    sum_val = (rounds * TEA_DELTA) & 0xFFFFFFFF
    y = v[0]

    while sum_val != 0:
        e = (sum_val >> 2) & 3

        for p in range(n - 1, 0, -1):
            z = v[p - 1]
            mx_val = _mx(y, z, sum_val, key[(p & 3) ^ e])
            v[p] = (v[p] - mx_val) & 0xFFFFFFFF
            y = v[p]

        z = v[n - 1]
        mx_val = _mx(y, z, sum_val, key[0 ^ e])
        v[0] = (v[0] - mx_val) & 0xFFFFFFFF
        y = v[0]

        sum_val = (sum_val - TEA_DELTA) & 0xFFFFFFFF

    return v

@njit(cache=True)
def _encrypt_block_fixed(v: np.ndarray, key: np.ndarray) -> np.ndarray:
    """JIT-compiled encryption logic for fixed 16-word blocks"""
    rounds = 6 + 52 // MAX_WORDS
    sum_val = np.uint32(0)
    z = v[MAX_WORDS - 1]

    for _ in range(rounds):
        sum_val = (sum_val + TEA_DELTA) & 0xFFFFFFFF
        e = (sum_val >> 2) & 3

        # Process all elements
        for p in range(MAX_WORDS - 1):
            y = v[p + 1]
            mx_val = _mx(y, z, sum_val, key[(p & 3) ^ e])
            v[p] = (v[p] + mx_val) & 0xFFFFFFFF
            z = v[p]

        y = v[0]
        mx_val = _mx(y, z, sum_val, key[((MAX_WORDS - 1) & 3) ^ e])
        v[MAX_WORDS - 1] = (v[MAX_WORDS - 1] + mx_val) & 0xFFFFFFFF
        z = v[MAX_WORDS - 1]

    return v

@njit(cache=True)
def _decrypt_block_fixed(v: np.ndarray, key: np.ndarray) -> np.ndarray:
    """JIT-compiled decryption logic for fixed 16-word blocks"""
    rounds = 6 + 52 // MAX_WORDS
    sum_val = (rounds * TEA_DELTA) & 0xFFFFFFFF
    y = v[0]

    while sum_val != 0:
        e = (sum_val >> 2) & 3

        # Process all elements
        for p in range(MAX_WORDS - 1, 0, -1):
            z = v[p - 1]
            mx_val = _mx(y, z, sum_val, key[(p & 3) ^ e])
            v[p] = (v[p] - mx_val) & 0xFFFFFFFF
            y = v[p]

        z = v[MAX_WORDS - 1]
        mx_val = _mx(y, z, sum_val, key[0 ^ e])
        v[0] = (v[0] - mx_val) & 0xFFFFFFFF
        y = v[0]

        sum_val = (sum_val - TEA_DELTA) & 0xFFFFFFFF

    return v

@njit(cache=True, parallel=True)
def _encrypt_blocks_parallel(data: np.ndarray, key: np.ndarray, block_count: int) -> np.ndarray:
    # Process blocks in parallel using prange
    for i in prange(block_count):
        # Extract block
        block_start = i * MAX_WORDS
        block_end = block_start + MAX_WORDS
        v = data[block_start:block_end].copy()

        # Encrypt block
        v = _encrypt_block_fixed(v, key)

        # Write back
        data[block_start:block_end] = v

    return data

@njit(cache=True, parallel=True)
def _decrypt_blocks_parallel(data: np.ndarray, key: np.ndarray, block_count: int) -> np.ndarray:
    # Process blocks in parallel using prange
    for i in prange(block_count):
        # Extract block
        block_start = i * MAX_WORDS
        block_end = block_start + MAX_WORDS
        v = data[block_start:block_end].copy()

        # Decrypt block
        v = _decrypt_block_fixed(v, key)

        # Write back
        data[block_start:block_end] = v

    return data




from .metadata import MetadataType
from typing import Callable, Dict
from hashlib import md5
from enum import Enum

# NOTE: osu!stream and osu! have different key generation methods, which can be found here:
#       https://github.com/ppy/osu-stream/blob/master/osu!stream/Helpers/osu!common/MapPackage.cs#L244

class KeyType(Enum):
    OSZ2 = "osz2"
    OSF2 = "osf2"

def generate_osz2_key(metadata: Dict[MetadataType, str]) -> bytes:
    assert MetadataType.Creator in metadata, "Metadata is missing creator"
    assert MetadataType.BeatmapSetID in metadata, "Metadata is missing beatmapset ID"

    creator = metadata[MetadataType.Creator]
    beatmapset_id = metadata[MetadataType.BeatmapSetID]
    seed = f"{creator}yhxyfjo5{beatmapset_id}"
    return md5(seed.encode("utf-8")).digest()

def generate_osf2_key(metadata: Dict[MetadataType, str]) -> bytes:
    assert MetadataType.Title in metadata, "Metadata is missing title"
    assert MetadataType.Artist in metadata, "Metadata is missing artist"
    
    title = metadata[MetadataType.Title]
    artist = metadata[MetadataType.Artist]
    seed = f"\x08{title}4390gn8931i{artist}"
    return md5(seed.encode("utf-8")).digest()

KeyGenerator = Callable[[Dict[MetadataType, str]], bytes]
Mapping: Dict[KeyType, KeyGenerator] = {
    KeyType.OSZ2: generate_osz2_key,
    KeyType.OSF2: generate_osf2_key,
}




from typing import Dict, List, Iterable, Optional
from .keys import KeyType, Mapping as KeyMapping
from .xxtea_reader import XXTEAReader
from .constants import KNOWN_PLAIN
from .metadata import MetadataType
from .file import File
from .xtea import XTEA
from .utils import *

import zipfile
import struct
import io

class Osz2Package:
    def __init__(
        self,
        reader: io.BufferedReader,
        metadata_only: bool = False,
        key_type: KeyType = KeyType.OSZ2
    ) -> None:
        self.metadata: Dict[MetadataType, str] = {}
        self.filenames: Dict[str, int] = {}
        self.files: List[File] = []
        self.key_type = key_type
        self.key: bytes = b""

        self.metadata_hash: bytes = b""
        self.file_info_hash: bytes = b""
        self.full_body_hash: bytes = b""

        # Always read the header when initializing
        self.read_header(reader)

        if not metadata_only:
            # Read the files if requested
            self.read_files(reader)

    @classmethod
    def from_file(cls, path: str, metadata_only=False, key_type=KeyType.OSZ2) -> "Osz2Package":
        with open(path, "rb") as f:
            return cls(f, metadata_only, key_type)

    @classmethod
    def from_bytes(cls, data: bytes, metadata_only=False, key_type=KeyType.OSZ2) -> "Osz2Package":
        with io.BytesIO(data) as f:
            return cls(f, metadata_only, key_type)

    @property
    def beatmap_files(self) -> Iterable[File]:
        for file in self.files:
            if file.is_beatmap:
                yield file

    @property
    def osz_filename(self) -> str:
        return sanitize_filename(
            f'{self.metadata.get(MetadataType.BeatmapSetID, "")} '
            f'{self.metadata.get(MetadataType.Artist, "Unknown")} - '
            f'{self.metadata.get(MetadataType.Title, "Unknown")} '
            f'({self.metadata.get(MetadataType.Creator, "Unknown")})'
        ).strip()  + '.osz'

    def find_file_by_name(self, name: str) -> Optional[File]:
        """Get a file by its filename"""
        return next((file for file in self.files if file.filename == name), None)

    def create_osz_package(
        self,
        compression: int = zipfile.ZIP_DEFLATED,
        exclude_disallowed_files: bool = True
    ) -> bytes:
        """Create a regular .osz package from the current files"""
        with io.BytesIO() as buffer:
            osz = zipfile.ZipFile(buffer, 'w', compression)

            for file in self.files:
                if exclude_disallowed_files and not file.is_allowed_extension:
                    # See `constants.ALLOWED_FILE_EXTENSIONS` for allowed file extensions
                    continue

                # Create ZipInfo to set file metadata
                zip_info = zipfile.ZipInfo(filename=file.filename)
                zip_info.compress_type = compression
                zip_info.date_time = file.date_modified.timetuple()[:6]
                osz.writestr(zip_info, file.content)

            osz.close()
            return buffer.getvalue()

    def calculate_osz_filesize(
        self,
        compression: int = zipfile.ZIP_DEFLATED,
        exclude_disallowed_files: bool = True
    ) -> int:
        """Calculate the size of the .osz package if it were to be created"""
        return len(self.create_osz_package(compression, exclude_disallowed_files))

    def read_header(self, reader: io.BufferedReader) -> None:
        magic = reader.read(3)
        assert magic == b"\xECHO", "Not a valid osz2 package" # nice one echo

        # Seek 17 bytes from the current position
        # to skip unused version byte & IV data
        reader.seek(17, 1)

        self.metadata_hash = reader.read(16)
        self.file_info_hash = reader.read(16)
        self.full_body_hash = reader.read(16)

        self.read_metadata(reader)
        self.read_file_names(reader)

        # Generate key based on metadata and key type
        # Usually this is just the MD5 of "<creator>yhxyfjo5<beatmapsetID>"
        key_generator = KeyMapping[self.key_type]
        self.key = key_generator(self.metadata)

    def read_metadata(self, reader: io.BufferedReader) -> None:
        buffer = reader.read(4)
        count = struct.unpack("<I", buffer)[0]

        for _ in range(count):
            buf = reader.read(2)
            meta_type = struct.unpack("<H", buf)[0]
            meta_value = read_string(reader)

            self.metadata[MetadataType(meta_type)] = meta_value

            buffer += buf
            buffer += write_string(meta_value)

        hash = compute_osz_hash(buffer, count*3, 0xA7)
        assert hash == self.metadata_hash, f"Metadata hash mismatch, expected: {hash}, got: {self.metadata_hash}"

    def read_file_names(self, reader: io.BufferedReader) -> None:
        buffer = reader.read(4)
        count = struct.unpack("<I", buffer)[0]

        for _ in range(count):
            filename = read_string(reader)
            beatmap_id = struct.unpack("<I", reader.read(4))[0]
            self.filenames[filename] = beatmap_id

    def read_files(self, reader: io.BufferedReader) -> None:
        # Convert key to uint32 array for XXTEA
        key = bytes_to_uint32_array(self.key)

        # Verify encrypted magic
        encrypted_magic = bytearray(reader.read(64))
        xtea = XTEA(key)
        xtea.decrypt(encrypted_magic, 0, 64)
        assert encrypted_magic == KNOWN_PLAIN, "Invalid encryption key"

        # Read encrypted i32 length
        length = struct.unpack("<I", reader.read(4))[0]

        # Decode length by encrypted length
        for i in range(0, 16, 2):
            length -= self.file_info_hash[i] | (self.file_info_hash[i+1] << 17)

        file_info = reader.read(length)
        file_data = reader.read()
        file_offset = reader.seek(0, 1)
        total_size = reader.seek(0, 2)
        reader.seek(file_offset, 0)

        # Parse file infos using xxtea stream
        with XXTEAReader(io.BytesIO(file_info), key) as xxtea:
            count = struct.unpack("<I", xxtea.read(4))[0]
            curr_offset = struct.unpack("<I", xxtea.read(4))[0]

            # Verify file info hash
            file_info_hash = compute_osz_hash(file_info, count*4, 0xd1)
            assert file_info_hash == self.file_info_hash, f"File info hash mismatch, expected: {file_info_hash}, got: {self.file_info_hash}"

            for i in range(count):
                filename = read_string(xxtea)
                file_hash = xxtea.read(16)

                date_created_binary = struct.unpack("<Q", xxtea.read(8))[0]
                date_modified_binary = struct.unpack("<Q", xxtea.read(8))[0]

                # Convert from .NET DateTime.ToBinary() format
                date_created = datetime_from_binary(date_created_binary)
                date_modified = datetime_from_binary(date_modified_binary)

                next_offset = total_size - file_offset
                if count > i + 1:
                    next_offset = struct.unpack("<I", xxtea.read(4))[0]

                file_length = next_offset - curr_offset
                file = File(
                    filename,
                    curr_offset,
                    file_length,
                    file_hash,
                    date_created,
                    date_modified,
                    content=bytes(),
                )
                self.files.append(file)
                curr_offset = next_offset

        # After reading the file info, read the actual file contents
        with XXTEAReader(io.BytesIO(file_data), key) as xxtea:
            for i in range(len(self.files)):
                length = struct.unpack("<I", xxtea.read(4))[0]
                self.files[i].content = xxtea.read(length)




from .constants import ALLOWED_FILE_EXTENSIONS
from dataclasses import dataclass
from datetime import datetime

@dataclass
class File:
    filename: str
    offset: int
    size: int
    hash: bytes
    date_created: datetime
    date_modified: datetime
    content: bytes

    @property
    def is_beatmap(self) -> bool:
        return self.filename.endswith('.osu')

    @property
    def file_extension(self) -> str:
        return (
            self.filename.split('.')[-1].lower()
            if '.' in self.filename else ''
        )

    @property
    def is_allowed_extension(self) -> bool:
        return self.file_extension in ALLOWED_FILE_EXTENSIONS




from osz2.xxtea import XXTEA
from typing import List
from io import BytesIO

class XXTEAReader:
    def __init__(self, reader: BytesIO, key: List[int]) -> None:
        self.reader: BytesIO = reader
        self.xxtea: XXTEA = XXTEA(key)

    def __enter__(self) -> "XXTEAReader":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.reader.close()

    def read(self, n: int) -> bytes:
        read = bytearray(self.reader.read(n))
        self.xxtea.decrypt(read, 0, n)
        return bytes(read)




from osz2.package import Osz2Package, KeyType
from pathlib import Path

import argparse
import sys
import os

def decrypt_osz2(filepath: str, key_type: KeyType) -> Osz2Package:
    if not os.path.exists(filepath):
        print(f"Error: Input file does not exist: {filepath}", file=sys.stderr)

    print("Reading osz2 package...")
    return Osz2Package.from_file(filepath, key_type=key_type)

def save_osz2(package: Osz2Package, output: str) -> None:
    Path(output).mkdir(exist_ok=True)
    print(f"Extracting {len(package.files)} files to {output}")

    for file in package.files:
        output_path = os.path.join(output, file.filename)

        if (dir := Path(output_path).parent) != ".":
            dir.mkdir(parents=True, exist_ok=True)

        with open(output_path, "wb") as f:
            f.write(file.content)

        print(f"  -> {file.filename} ({len(file.content)} bytes)")

def main() -> None:
    parser = argparse.ArgumentParser(prog="osz2", description="A tool to decrypt and extract osz2 files")
    parser.add_argument("input", help="The path to the osz2 file to decrypt (required)")
    parser.add_argument("output", help="The path to put the extracted osz2 files (required)")
    parser.add_argument("--key-type", choices=["osz2", "osf2"], default="osz2", help="The key generation method to use (default: osz2)")
    parser.add_argument("--create-osz", action="store_true", help="Create a regular .osz package from the decrypted files")
    args = parser.parse_args()
    key_type = KeyType(args.key_type)

    osz2 = decrypt_osz2(args.input, key_type)
    save_osz2(osz2, args.output)

    if args.create_osz:
        osz_data = osz2.create_osz_package(exclude_disallowed_files=False)

        with open(f"{args.output}/{osz2.osz_filename}", "wb") as f:
            f.write(osz_data)

if __name__ == "__main__":
    main()




# "knownPlain" constant derived from FastRandom(1990)
# https://github.com/ppy/osu-stream/blob/master/osu!stream/Helpers/osu!common/MapPackage.cs#L64
KNOWN_PLAIN = bytearray([
   0x55, 0xAA, 0x74, 0x10, 0x2B, 0x56, 0xB3, 0x9E,
   0x25, 0x9E, 0xFE, 0xB7, 0xBE, 0x06, 0xFC, 0xF2,
   0xB6, 0x3C, 0x6F, 0x47, 0x7E, 0x38, 0x69, 0x43,
   0x80, 0x89, 0x25, 0x00, 0xCC, 0xB6, 0xFE, 0x12,
   0xA9, 0xB2, 0x4A, 0x2C, 0x96, 0xD5, 0xEA, 0x26,
   0x42, 0x31, 0xAF, 0x0A, 0x0D, 0xAE, 0x00, 0xED,
   0xFE, 0x96, 0xA6, 0x94, 0x99, 0xA7, 0x90, 0xE4,
   0x68, 0xBF, 0xC6, 0x97, 0x5B, 0x1B, 0x5E, 0x7F
])

# A list of all allowed file extensions in an .osz package
# Did not cause any issues when testing on titanic
ALLOWED_FILE_EXTENSIONS = (
    "osu", "osz", "osb", "osk", "png", "mp3",
    "wav", "ogg", "jpg", "wmv", "flv", "flac",
    "avi", "ini",  "m4v", "mpg", "mov", "webm",
    "ogv", "mpeg", "3gp", "mkv", "mp4", "jpeg",
)




__author__ = "Lekuru"
__email__ = "contact@lekuru.xyz"
__version__ = "1.0.5"
__license__ = "MIT"

from .patch import apply_bsdiff_patch
from .metadata import MetadataType
from .package import Osz2Package
from .keys import KeyType
from .file import File

from .simple_cryptor import SimpleCryptor
from .xxtea_reader import XXTEAReader
from .xxtea import XXTEA
from .xtea import XTEA




import datetime
import hashlib
import struct
import typing
import io
import re

unsafe_characters_pattern = re.compile(r'[<>:"/\\|?*\x00-\x1F]')

def sanitize_filename(filename: str) -> str:
    return re.sub(unsafe_characters_pattern, "", filename)

def bytes_to_uint32_array(data: bytes) -> typing.List[int]:
    return [x[0] for x in struct.iter_unpack("<I", data)]

def uint32_slice_to_byte_slice(u32s: typing.List[int]) -> typing.List[int]:
    bytes_list = []
    for u32 in u32s:
        bytes_list.append(u32 & 0xFF)
        bytes_list.append((u32 >> 8) & 0xFF)
        bytes_list.append((u32 >> 16) & 0xFF)
        bytes_list.append((u32 >> 24) & 0xFF)
    return bytes_list

def read_string(reader: io.BufferedReader) -> str:
    length = read_uleb128(reader)
    if length == 0:
        return ""
    return reader.read(length).decode("utf-8")

def write_string(string: str) -> bytes:
    encoded = string.encode("utf-8")
    buf = write_uleb128(len(encoded))
    return buf + encoded

def read_uleb128(reader: io.BufferedReader) -> int:
    result = 0
    shift = 0

    while True:
        b = reader.read(1)
        if not b:
            raise EOFError("Unexpected end of file while reading ULEB128")

        byte = b[0]
        result |= (byte & 0x7F) << shift

        if (byte & 0x80) == 0:
            break

        shift += 7

    return result

def write_uleb128(value: int) -> bytes:
    buf = bytearray()

    while True:
        byte = value & 0x7F
        value >>= 7
        if value != 0:
            byte |= 0x80

        buf.append(byte)

        if value == 0:
            break

    return bytes(buf)

def compute_osz_hash(buffer: bytes, pos: int, swap: int) -> bytes:
    buf = bytearray(buffer)

    if pos < 0 or pos >= len(buf):
        # If pos is out of bounds, just compute hash without swapping
        hash_bytes = bytearray(hashlib.md5(buf).digest())
    else:
        buf[pos] ^= swap
        hash_bytes = bytearray(hashlib.md5(buf).digest())
        buf[pos] ^= swap # restore original

    # Swap bytes as in C# implementation
    for i in range(8):
        hash_bytes[i], hash_bytes[i+8] = hash_bytes[i+8], hash_bytes[i]

    hash_bytes[5] ^= 0x2D
    return bytes(hash_bytes)

def datetime_from_binary(time: int) -> datetime.datetime:
    n_ticks = time & 0x3FFFFFFFFFFFFFFF
    secs = n_ticks / 1e7
    d1 = datetime.datetime(1, 1, 1)
    t1 = datetime.timedelta(seconds=secs)
    return d1 + t1




import bsdiff4
import gzip
import io

def apply_bsdiff_patch(source_osz2: bytes, patch_bytes: bytes) -> bytes:
    """Apply a BSDIFF4-format patch to an osz2 package."""
    return bsdiff4.core.patch(
        source_osz2,
        *read_gzip_patch(patch_bytes)
    )

def read_gzip_patch(patch_bytes: bytes, compressor=gzip) -> tuple:
    """
    Read a BSDIFF4-format patch from bytes 'patch_bytes'
    with control over the compression algorithm.
    (osu! uses gzip compression for its patches)
    """
    fi = io.BytesIO(patch_bytes)
    magic = fi.read(8)

    if magic[:7] != b'BSDIFF40'[:7]:
        raise ValueError("incorrect magic bsdiff4 header")

    # length headers
    len_control = bsdiff4.core.decode_int64(fi.read(8))
    len_diff = bsdiff4.core.decode_int64(fi.read(8))
    len_dst = bsdiff4.core.decode_int64(fi.read(8))

    # read the control header
    bcontrol = compressor.decompress(fi.read(len_control))
    tcontrol = [
        (bsdiff4.core.decode_int64(bcontrol[i:i + 8]),
         bsdiff4.core.decode_int64(bcontrol[i + 8:i + 16]),
         bsdiff4.core.decode_int64(bcontrol[i + 16:i + 24]))
         for i in range(0, len(bcontrol), 24)
    ]

    # read the diff and extra blocks
    bdiff = compressor.decompress(fi.read(len_diff))
    bextra = compressor.decompress(fi.read())
    return len_dst, tcontrol, bdiff, bextra
